/*
 * clisp2.c - CAMPUS LIsP, Mint version
 * by Hirotsugu Kakugawa
 * Based on a lisp interpreter for VFlib version 4.
 *
 * 12 Sep 2000
 * 18 Sep 2000  Added vector and string.
 * 30 Oct 2002  CAMPUS LIsP, Mint version
 */
/*
 * Copyright (C) 2000, 2002  Hirotsugu Kakugawa. 
 * All rights reserved.
 *
 * This file is part of the VFlib Library.  This library is free
 * software; you can redistribute it and/or modify it under the terms of
 * the GNU Library General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  This library is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU Library General Public License for more details.
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef CAMPUS_LISP
#include "../../config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#endif
#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif
#include <ctype.h>
#include <sys/types.h>
#include <sys/param.h>

#ifndef CAMPUS_LISP
#include "VFlib40.h"
#include "vflib.h"
#define  malloc(n)    vf_malloc(NULL,n)
#define  calloc(n,s)  vf_calloc(NULL,n,s)
#define  realloc(p,n) vf_realloc(NULL,p,n)
#define  free(p)      vf_free(NULL,p)
#endif /*CAMPUS_LISP*/


/* Data Representation ('int' must be at least 32 bits) */
typedef int    lobj;
#define D_MASK_DATA     0x03ffffff
#define D_MASK_TAG      0x78000000
#define D_GC_MARK       0x80000000
#define D_TAG_BIT_POS   27
#define D_INT_SIGN_BIT  0x02000000
#define D_GET_TAG(s)    (s & ~(D_GC_MARK | D_MASK_DATA))
#define D_GET_DATA(s)   (s & D_MASK_DATA)

/* Data Tags */
#define TAG_NIL     (0 << D_TAG_BIT_POS)
#define TAG_T       (1 << D_TAG_BIT_POS)
#define TAG_EOF     (2 << D_TAG_BIT_POS)
#define TAG_UNDEF   (3 << D_TAG_BIT_POS)
#define TAG_INT     (4 << D_TAG_BIT_POS)
#define TAG_SYMB    (5 << D_TAG_BIT_POS)
#define TAG_CONS    (6 << D_TAG_BIT_POS)
#define TAG_VECT    (7 << D_TAG_BIT_POS)
#define TAG_STR     (8 << D_TAG_BIT_POS)
#define TAG_FLOAT   (9 << D_TAG_BIT_POS)

#define NCONS      (1024*8)
#define NVECTS       256
#define NSTRS        256
#define NFLOATS      256
#define NSYMBS       256
#define PNAME_SIZE  1024
#define STACK_SIZE  1024

typedef struct s_lisp  *LISP;
struct  s_lisp {
  /* Cells */
  int    t_cons_ncons;
  int    t_cons_free;           /* free list */
  lobj  *t_cons_car;            /* "car" part of cell */
  lobj  *t_cons_cdr;            /* "cdr" part of cell */
  /* Vectors */
  int    t_vect_nvects;
  int    t_vect_free;           /* free list */
  int   *t_vect_size;           /* array for vector size */
  lobj **t_vect_body;           /* array for vector body */
  char  *t_vect_gcmark;         /* marks for gc */
  /* Strings */
  int    t_str_nstrs;
  int    t_str_free;            /* free list */
  int   *t_str_size;            /* array for string size */
  char **t_str_body;            /* array for string ptr */
  char  *t_str_gcmark;          /* marks for gc */
  /* Floats */
  int    t_flo_nfloats;
  int    t_flo_free;            /* free list */
  union {
    double   v;                 /* array for floats */
    int      next;
  } *t_flo_body;                /* array for floats */
  char  *t_flo_gcmark;          /* marks for gc */
  /* Symbols */
  int    t_symb_nsymbs;
  int    t_symb_free;           /* free slot */
  int   *t_symb_pname;          /* printable name (index of t_pnames[]) */
  lobj  *t_symb_val;            /* symbol value */
  lobj  *t_symb_fval;           /* symbol function definition */
  int   *t_symb_ftype;          /* function type */ 
  /* Printable name */
  int    t_pnames_size;
  int    t_pnames_free;         /* free pointer */
  char  *t_pnames;              /* name buffer */ 
  /* Stack */
  int    t_stack_size;
  lobj  *t_stack;               /* the stack */
  int    t_stack_ptr;           /* stack pointer */
#ifndef CAMPUS_LISP
  vf_table   l_extern_gc_protect;
#endif
};


/* Function types */
enum Ftype {
  FTYPE_UNDEF,
  FTYPE_SYS, 
  FTYPE_SPECIAL,
  FTYPE_USER
};
#define FTYPE(t, nargs)       ((t)*1000 + (nargs))
#define FTYPE_ANY_ARGS         999
#define FTYPE_GET_TYPE(p)     ((p) / 1000)
#define FTYPE_GET_NARGS(p)    ((p) % 1000) 

/* Keywords */
enum keywords {
  KW_READ,     KW_EVAL,     KW_PRINT,    KW_TERPRI,
  KW_DEFUN,    KW_LAMBDA,   KW_QUOTE,    KW_SETQ,     KW_COMMENT,
  KW_QUIT,     KW_GC,       KW_READ_STR,
  KW_CONS,     KW_CAR,      KW_CDR,      KW_RPLACA,   KW_RPLACD,  
  KW_EQ,       KW_NULL,     KW_SYMBP,    
  KW_NUMBERP,  KW_INTEGERP, KW_FLOATP,   KW_EOFP,    
  KW_PROGN,    KW_COND,     KW_IF,       KW_AND,      KW_OR,    
  KW_NOT,      KW_LIST,     KW_EQUAL,    
  KW_ADD,      KW_SUB,      KW_TIMES,    KW_QUOTIENT, KW_DIVIDE,
  KW_IEQ,      KW_IGT,      KW_IGE,      KW_ILT,      KW_ILE,
  KW_VECTP,    KW_VECT,     KW_MKVECT,   KW_VECTLEN,  KW_VECTREF,  KW_VECTSET,
  KW_STRP,     KW_MKSTR,    KW_STRLEN,   KW_STRAPPND
};
struct s_keywords {
  char  *key;
  int    ftype;
  int    i;
};

/* Built-in function table */
static struct s_keywords funcs[] = {
  { "read",           FTYPE(FTYPE_SYS,     0),               KW_READ     },
  { "eval",           FTYPE(FTYPE_SYS,     1),               KW_EVAL     },
  { "print",          FTYPE(FTYPE_SYS,     1),               KW_PRINT    },
  { "terpri",         FTYPE(FTYPE_SYS,     0),               KW_TERPRI   },
  { "defun",          FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_DEFUN    },
  { "lambda",         FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_LAMBDA   },
  { "quote",          FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_QUOTE    },
  { "setq",           FTYPE(FTYPE_SPECIAL, 2),               KW_SETQ     },
  { "comment",        FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_COMMENT  },
  { "quit",           FTYPE(FTYPE_SYS,     0),               KW_QUIT     },
  { "gc",             FTYPE(FTYPE_SYS,     0),               KW_GC       },
  { "read-string",    FTYPE(FTYPE_SYS,     1),               KW_READ_STR },
  { "cons",           FTYPE(FTYPE_SYS,     2),               KW_CONS     },
  { "car",            FTYPE(FTYPE_SYS,     1),               KW_CAR      },
  { "cdr",            FTYPE(FTYPE_SYS,     1),               KW_CDR      },
  { "rplaca",         FTYPE(FTYPE_SYS,     2),               KW_RPLACA   },
  { "rplacd",         FTYPE(FTYPE_SYS,     2),               KW_RPLACD   },
  { "eq",             FTYPE(FTYPE_SYS,     2),               KW_EQ       },
  { "null",           FTYPE(FTYPE_SYS,     1),               KW_NULL     },
  { "symbolp",        FTYPE(FTYPE_SYS,     1),               KW_SYMBP    },
  { "numberp",        FTYPE(FTYPE_SYS,     1),               KW_NUMBERP  },
  { "integerp",       FTYPE(FTYPE_SYS,     1),               KW_INTEGERP },
  { "floatp",         FTYPE(FTYPE_SYS,     1),               KW_FLOATP   },
  { "eofp",           FTYPE(FTYPE_SYS,     1),               KW_EOFP     },
  { "progn",          FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_PROGN    },
  { "cond",           FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_COND     },
  { "if",             FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_IF       },
  { "and",            FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_AND      },
  { "or",             FTYPE(FTYPE_SPECIAL, FTYPE_ANY_ARGS),  KW_OR       },
  { "not",            FTYPE(FTYPE_SYS,     1),               KW_NOT      },
  { "list",           FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_LIST     },
  { "equal",          FTYPE(FTYPE_SYS,     2),               KW_EQUAL    },
  { "+",              FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_ADD      },
  { "-",              FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_SUB      },
  { "*",              FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_TIMES    },
  { "/",              FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_QUOTIENT },
  { "divide",         FTYPE(FTYPE_SYS,     2),               KW_DIVIDE   },
  { "=",              FTYPE(FTYPE_SYS,     2),               KW_IEQ      },
  { ">",              FTYPE(FTYPE_SYS,     2),               KW_IGT      },
  { ">=",             FTYPE(FTYPE_SYS,     2),               KW_IGE      },
  { "<",              FTYPE(FTYPE_SYS,     2),               KW_ILT      },
  { "<=",             FTYPE(FTYPE_SYS,     2),               KW_ILE      },
  { "vectorp",        FTYPE(FTYPE_SYS,     1),               KW_VECTP    },
  { "vector",         FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_VECT     },
  { "make-vector",    FTYPE(FTYPE_SYS,     1),               KW_MKVECT   },
  { "vector-length",  FTYPE(FTYPE_SYS,     1),               KW_VECTLEN  },
  { "vector-ref",     FTYPE(FTYPE_SYS,     2),               KW_VECTREF  },
  { "vector-set",     FTYPE(FTYPE_SYS,     3),               KW_VECTSET  },
  { "stringp",        FTYPE(FTYPE_SYS,     1),               KW_STRP     },
  { "make-string",    FTYPE(FTYPE_SYS,     1),               KW_MKSTR    },
  { "string-length",  FTYPE(FTYPE_SYS,     1),               KW_STRLEN   },
  { "string-append",  FTYPE(FTYPE_SYS,     FTYPE_ANY_ARGS),  KW_STRAPPND },
  { NULL,             -1,                                    -1          }
};

/* Error messages */
static char *errmsg_sym_undef  = "SYMBOL UNDEFINED: ";
static char *errmsg_func_undef = "FUNCTION UNDEFINED: ";
static char *errmsg_ill_nargs  = "ILLEGAL NUMBER OF ARGUMENTS: ";
static char *errmsg_ill_type   = "ILLEGAL ARGUMENT TYPE: ";
static char *errmsg_ill_call   = "ILLEGAL FUNCTION CALL: ";
static char *errmsg_ill_syntax = "ILLEGAL SYNTAX: ";
static char *errmsg_ill_sexp   = "ILLEGAL S-EXP: ";
static char *errmsg_ill_range  = "ARGUMENT OUT OF RANGE: ";
static char *errmsg_eof        = "END OF FILE";
static char *errmsg_stack_of   = "STACK OVERFLOW";
static char *errmsg_zero_div   = "DIVISION BY ZERO: ";

/* Function types */
static void   init(LISP);
static void   toplevel(LISP);
static lobj   l_read(LISP), l_eval(LISP,lobj), l_print(LISP,lobj);
static lobj   l_read_fp(LISP,FILE*);
static int    getc_fp(void*,void*);
static void   ungetc_fp(int,void*,void*);
static lobj   l_read_string(LISP,char*);
static int    getc_string(void*,void*);
static void   ungetc_string(int,void*,void*);
static lobj   l_read_stream(LISP,int(*g)(),void(*u)(),void*,void*);
static int    skip_space(int(*g)(),void(*u)(),void*,void*);
static lobj   int_make_l(int v);
static int    int_get_c(lobj);
static double float_get_c(LISP,lobj);
static lobj   eval_args(LISP,lobj,lobj,lobj av[3],int);
static lobj   special(LISP,lobj,lobj);
static lobj   fcall(LISP,lobj,lobj av[3],int);
static lobj   apply(LISP,lobj,lobj,int);
static int    err_msg(LISP,char*,int,lobj);
static lobj   l_cons(LISP,lobj,lobj);
static lobj   l_car(LISP,lobj), l_cdr(LISP,lobj);
static lobj   l_make_vect(LISP,int), l_list_to_vect(LISP,lobj);
static lobj   l_make_str(LISP,int), l_make_float(LISP,double);
static lobj   l_equal(LISP,lobj,lobj);
static void   quit(LISP);
static int    list_len(LISP,lobj);
static void   rplacd(LISP,lobj,lobj), rplaca(LISP,lobj,lobj);
static void   gcollect(LISP);
static void   gc_mark(LISP,lobj);
static int    gc_protect(LISP,lobj);
static void   gc_unprotect(LISP,lobj);
static lobj   symb_make(LISP,char*);




#ifdef CAMPUS_LISP
int
main(int argc, char **argv)
{
  struct s_lisp  lisp;

  init(&lisp);
  printf("CAMPUS LIsP 2.0, Mint version\n");
  toplevel(&lisp);
  quit(&lisp);
  
  return 0;
}
#endif


/* Initialize lisp strage */
static void
init(LISP l)
{
  int   i;

  /* cells */
  l->t_cons_ncons = NCONS;
  l->t_cons_car = (lobj*) calloc(l->t_cons_ncons, sizeof(lobj));
  l->t_cons_cdr = (lobj*) calloc(l->t_cons_ncons, sizeof(lobj));
  /* make a free cell list */
  l->t_cons_free = 0;
  for (i = 0; i < l->t_cons_ncons-1; i++)
    l->t_cons_car[i] = i+1;
  l->t_cons_car[l->t_cons_ncons-1] = l->t_cons_ncons-1;

  /* vectors */
  l->t_vect_nvects = NVECTS;
  l->t_vect_size   = (int*)  calloc(l->t_vect_nvects, sizeof(int));
  l->t_vect_body   = (lobj**)calloc(l->t_vect_nvects, sizeof(lobj));
  l->t_vect_gcmark = (char*) calloc(l->t_vect_nvects, sizeof(char));
  /* make free list of vectors */
  l->t_vect_free = 0;
  for (i = 0; i < l->t_vect_nvects-1; i++)
    l->t_vect_size[i] = i+1;
  l->t_vect_size[l->t_vect_nvects-1] = l->t_vect_nvects-1;
  for (i = 0; i < l->t_vect_nvects; i++){
    l->t_vect_body[i]   = NULL;
    l->t_vect_gcmark[i] = 0;
  }

  /* strings */
  l->t_str_nstrs = NSTRS;
  l->t_str_size   = (int*)   calloc(l->t_str_nstrs, sizeof(int));
  l->t_str_body   = (char**) calloc(l->t_str_nstrs, sizeof(char*));
  l->t_str_gcmark = (char*)  calloc(l->t_str_nstrs, sizeof(char));
  /* make free list of strings */
  l->t_str_free = 0;
  for (i = 0; i < l->t_str_nstrs-1; i++)
    l->t_str_size[i] = i+1;
  l->t_str_size[l->t_str_nstrs-1] = l->t_str_nstrs-1;
  for (i = 0; i < l->t_str_nstrs; i++){
    l->t_str_body[i]   = NULL;
    l->t_str_gcmark[i] = 0;
  }

  /* floats */
  l->t_flo_nfloats = NFLOATS;
  l->t_flo_body   = calloc(l->t_flo_nfloats, sizeof(double));
  l->t_flo_gcmark = (char*)  calloc(l->t_flo_nfloats, sizeof(char));
  /* make free list of strings */
  l->t_flo_free = 0;
  for (i = 0; i < l->t_flo_nfloats-1; i++)
    l->t_flo_body[i].next = i+1;
  l->t_flo_body[l->t_flo_nfloats-1].next = l->t_flo_nfloats-1;
  for (i = 0; i < l->t_flo_nfloats; i++)
    l->t_flo_gcmark[i] = 0;

  /* symbol table */
  l->t_symb_nsymbs = NSYMBS;
  l->t_symb_pname = (int*)  calloc(l->t_symb_nsymbs, sizeof(int));
  l->t_symb_val   = (lobj*) calloc(l->t_symb_nsymbs, sizeof(lobj));
  l->t_symb_fval  = (lobj*) calloc(l->t_symb_nsymbs, sizeof(lobj));
  l->t_symb_ftype = (int*)  calloc(l->t_symb_nsymbs, sizeof(int));
  l->t_symb_free = 0;
  for (i = 0; i < l->t_symb_nsymbs; i++)
    l->t_symb_pname[i] = -1;

  /* print name (symbol name) */
  l->t_pnames_size = PNAME_SIZE;
  l->t_pnames = (char*) malloc(l->t_pnames_size);
  l->t_pnames_free = 0;

  /* stack */
  l->t_stack_size = STACK_SIZE;
  l->t_stack = (lobj*) calloc(l->t_stack_size, sizeof(lobj));  
  l->t_stack_ptr = 0;

  /* install built-in functions */
  for (i = 0; funcs[i].key != NULL; i++){
    symb_make(l, funcs[i].key);
    l->t_symb_ftype[i] = funcs[i].ftype;
    if (i != funcs[i].i){
      printf("function install error: %s\n", funcs[i].key);
      quit(l);
    }
  }
}

#ifndef CAMPUS_LISP
static void
finish(LISP l)
{
  int  i;

  vf_free(NULL, l->t_cons_car); 
  vf_free(NULL, l->t_cons_cdr); 
  l->t_cons_ncons = 0;
  l->t_cons_free  = 0;
  l->t_cons_car = NULL;
  l->t_cons_cdr = NULL;

  if (l->t_vect_body != NULL){
    for (i = 0; i < l->t_vect_nvects; i++)
      vf_free(NULL, l->t_vect_body[i]);
  }
  vf_free(NULL, l->t_vect_size);
  vf_free(NULL, l->t_vect_body);
  vf_free(NULL, l->t_vect_gcmark);
  l->t_vect_nvects = 0;
  l->t_vect_free = 0;
  l->t_vect_size   = NULL;
  l->t_vect_body   = NULL;
  l->t_vect_gcmark = NULL;

  if (l->t_str_body != NULL){
    for (i = 0; i < l->t_str_nstrs; i++)
      vf_free(NULL, l->t_str_body[i]);
  }
  vf_free(NULL, l->t_str_size);
  vf_free(NULL, l->t_str_body);
  vf_free(NULL, l->t_str_gcmark);
  l->t_str_nstrs  = 0;
  l->t_str_free   = 0;
  l->t_str_size   = NULL;
  l->t_str_body   = NULL;
  l->t_str_gcmark = NULL;

  l->t_flo_nfloats = 0;
  l->t_flo_free    = 0;
  vf_free(NULL, l->t_flo_body);
  l->t_flo_body    = 0;
  vf_free(NULL, l->t_flo_gcmark);
  l->t_flo_gcmark  = 0;

  vf_free(NULL, l->t_symb_pname);
  vf_free(NULL, l->t_symb_val);
  vf_free(NULL, l->t_symb_fval);
  vf_free(NULL, l->t_symb_ftype);
  l->t_symb_nsymbs = 0;
  l->t_symb_free   = 0;
  l->t_symb_pname  = NULL;
  l->t_symb_val    = NULL;
  l->t_symb_fval   = NULL;
  l->t_symb_ftype  = NULL;

  vf_free(NULL, l->t_pnames);
  l->t_pnames_size = 0;
  l->t_pnames_free = 0;
  l->t_pnames      = NULL;

  vf_free(NULL, l->t_stack);
  l->t_stack_size = 0;
  l->t_stack_ptr  = 0;
  l->t_stack      = NULL;
}
#endif

/* Top level */
static void
toplevel(LISP l)
{
  lobj  s, v;

  for (;;){
    l->t_stack_ptr = 0;
    printf("\n] ");              /* prompt */
    if ((s = l_read(l)) < 0)     /* read */
      continue;
    if (s == TAG_EOF)            /* end of file */
      break;
    if (gc_protect(l, s) < 0)
      break;
    if ((v = l_eval(l, s)) < 0)  /* eval */
      continue;
    gc_unprotect(l, s);
    (void) l_print(l,v);         /* print */
  }
}


/* Read an S-expression */
static lobj
l_read(LISP l)
{
  return  l_read_fp(l, stdin);
}

static lobj
l_read_fp(LISP l, FILE *fp)
{
  int  uch = -1;

  return l_read_stream(l, getc_fp, ungetc_fp, (void*)fp, &uch); 
}

static int
getc_fp(void *p1, void *p2)
{
  int ch;
  int *uchp = (int*)p2;
  
  if (*uchp < 0)
    return getc((FILE*)p1);
  ch = *uchp;
  *uchp = -1;
  return ch;
}
static void
ungetc_fp(int ch, void *p1, void *p2)
{
  int *uchp = (int*)p2;

  *uchp = ch;
}

static lobj
l_read_string(LISP l, char *str)
{
  int  pos = 0;
  
  return l_read_stream(l, getc_string, ungetc_string, 
		       (void*)str, (void*)&pos); 
}

static int
getc_string(void *p1, void *p2)
{
  char *str = (char*)p1;
  int  *posp = (int*)p2;
  int   ch;

  ch = str[*posp];
  *posp = *posp + 1;
  if (ch == '\0')
    ch = EOF;
  return  ch;
}
static void
ungetc_string(int ch, void *p1, void *p2)
{
  int  *posp = (int*)p2;

  *posp = *posp - 1;
}

/* S-exp reader */
static lobj
l_read_stream(LISP l, int(*get_char)(), void(*unget_char)(), 
	      void *p1, void *p2)
{
  lobj  s, v, t;
  char  token[256];
  int   ch, i, j;
  
  /* skip spaces and comment */
  if ((ch = skip_space(get_char,unget_char,p1,p2)) < 0)    /* eof */
    return TAG_EOF; 

  if (ch == '\''){        /* quote macro */
    if ((t = l_read_stream(l, get_char,unget_char,p1,p2)) < 0)
      return -1;
    if (t == TAG_EOF)
      return err_msg(l, errmsg_eof, 0, 0);
    t = l_cons(l, t, TAG_NIL);
    s = l_cons(l, (TAG_SYMB|KW_QUOTE), t);
    return s;

  } else if (ch == '#'){         /* vector */
    if ((ch = skip_space(get_char,unget_char,p1,p2)) < 0)
      return err_msg(l, errmsg_eof, 0, 0);
    if (ch != '(')
      return err_msg(l, errmsg_ill_sexp, 0, 0);
    unget_char(ch,p1,p2);
    if ((t = l_read_stream(l, get_char,unget_char,p1,p2)) < 0)
      return err_msg(l, errmsg_eof, 0, 0);
    if (t == TAG_EOF)
      return -1;
    s = l_list_to_vect(l, t);
    return s;

  } else if (ch == '"'){         /* string */
    int   pn, i;
    char *p;

    pn = 16;
    p = (char*) malloc(pn);
    for (i = 0; ; i++){
      if ((ch = get_char(p1,p2)) < 0)
	return err_msg(l, errmsg_eof, 0, 0);
      if (ch == '\"')
	break;
      switch (ch){
      default:   break;
      case '\\': 
	if ((ch = get_char(p1,p2)) < 0)
	  return err_msg(l, errmsg_eof, 0, 0);
	switch (ch){
	default: break;
	case 'a':  ch = '\a'; break;
	case 'b':  ch = '\b'; break;
	case 'f':  ch = '\f'; break;
	case 'n':  ch = '\n'; break;
	case 'r':  ch = '\r'; break;
	case 't':  ch = '\t'; break;
	case 'v':  ch = '\v'; break;
	}
	break;
      }
      if (i == pn){
	pn += 16;
	p = (char*) realloc(p, pn);
      }
      p[i] = ch;
    }
    s = l_make_str(l, i);
    for (j = 0; j < i; j++)
      l->t_str_body[D_GET_DATA(s)][j] = p[j];
    free(p);
    return s;

  } else if (ch != '('){         /* t, nil, symbol, integer, or float */
    token[0] = ch;
    for (i = 1; ; i++){
      ch = get_char(p1,p2);
      if (isspace(ch) || iscntrl(ch) || (ch < 0) 
	  || (ch == ';') || (ch == '(') || (ch == ')')){
	unget_char(ch,p1,p2);
	token[i] = '\0';
	if (isdigit((int)token[0]) 
	    || ((token[0] == '-') && isdigit((int)token[1]))
	    || ((token[0] == '+') && isdigit((int)token[1]))){   /* number */
	  if (strchr(token, '.') == NULL)
	    s = int_make_l(atoi(token));
	  else
	    s = l_make_float(l, atof(token));
	} else if (strcmp(token, "nil") == 0){                   /* nil */ 
	  s = TAG_NIL;
	} else if (strcmp(token, "t") == 0){                     /* t */
	  s = TAG_T;
	} else {                                                 /* symbol */
	  s = TAG_SYMB | symb_make(l, token);
	}
	break;
      }
      token[i] = ch;
    }
    return s;

  } else /* ch == '(' */ {       /* list */
    if ((ch = skip_space(get_char,unget_char,p1,p2)) < 0){
      return err_msg(l, errmsg_eof, 0, 0);
    } else if (ch == ')'){
      s = TAG_NIL;  /* "()" = nil */
    } else {
      unget_char(ch,p1,p2);
      if ((t = l_read_stream(l, get_char,unget_char,p1,p2)) < 0)
	return err_msg(l, errmsg_eof, 0, 0);
      if (t == TAG_EOF)
	return -1;
      if ((s = v = l_cons(l, t, TAG_NIL)) < 0)
	return -1;
      if (gc_protect(l, s) < 0)
	return -1;
      for (;;){
	if ((ch = skip_space(get_char,unget_char,p1,p2)) < 0)  /* look ahead */
	  return err_msg(l, errmsg_eof, 0, 0);
	if (ch == ')')
	  break;
	unget_char(ch,p1,p2);
	if ((t = l_read_stream(l, get_char,unget_char,p1,p2)) < 0)
	  return -1;
	if (t == TAG_EOF)
	  return err_msg(l, errmsg_eof, 0, 0);
	if ((t = l_cons(l, t, TAG_NIL)) < 0) 
	  return -1;
	rplacd(l, v, t);
	v = l_cdr(l, v);
      }
      gc_unprotect(l, s);
    }
    return s;

  }

  fprintf(stderr, "cannot happen: l_read_stream()\n");
  abort();
  return TAG_EOF;
}

static int
skip_space(int(*get_char)(), void(*unget_char)(), void *p1, void *p2)
{
  int  ch;

  for (;;){
    if ((ch = get_char(p1,p2)) < 0)
      return -1;     /* end-of-file */
    if ((ch != ';') && !isspace(ch) && !iscntrl(ch)){
      break;
    } else if (ch == ';'){    /* comment */
      while (ch != '\n')
	ch = get_char(p1,p2);
    }
  }
  return ch;
}


/* Print an S-expression */ 
static lobj
l_print(LISP l, lobj s)
{
  lobj  v, t;
  int   i, d, c;

  switch(D_GET_TAG(s)){

  case TAG_NIL:
    printf("nil");
    break;

  case TAG_T:
    printf("t");
    break;

  case TAG_INT:
    v = int_get_c(s);
    printf("%d", v);
    break;

  case TAG_FLOAT:
    printf("%f", float_get_c(l, s));
    break;

  case TAG_SYMB:
    i = s & D_MASK_DATA;
    printf("%s", &l->t_pnames[l->t_symb_pname[i]]);
    break;

  case TAG_EOF:
    printf("<eof>");
    break;

  case TAG_UNDEF:  /* for debugging */
    printf("<undefined>"); 
    break;

  case TAG_CONS:
    printf("(");
    t = s;
    l_print(l, l_car(l,t));
    while (D_GET_TAG(l_cdr(l,t)) == TAG_CONS) {
      printf(" ");
      t = l_cdr(l,t);
      l_print(l, l_car(l,t));
    }
    if (D_GET_TAG(l_cdr(l,t)) != TAG_NIL){
      printf(" . ");
      l_print(l, l_cdr(l,t)); 
    }
    printf(")");
    break;

  case TAG_VECT:
    printf("#(");
    d = D_GET_DATA(s);
    for (i = 0; i < l->t_vect_size[d]; i++){
      l_print(l, l->t_vect_body[d][i]);
      if (i < l->t_vect_size[d]-1)
	printf(" ");
    }
    printf(")");
    break;

  case TAG_STR:
    printf("\"");
    d = D_GET_DATA(s);
    for (i = 0; i < l->t_str_size[d]; i++){
      c = l->t_str_body[d][i];
      if ((0x20 <= c) && (c <= 0x7e)){
	switch (c){
	case '\'':   printf("\\\'");  break;
	case '"':    printf("\\\"");  break;
	case '\\':   printf("\\\\");  break;
	default:     printf("%c", c); break;
	}
      } else {
	switch (c){
	case '\a':   printf("\\a"); break;
	case '\b':   printf("\\b"); break;
	case '\f':   printf("\\f"); break;
	case '\n':   printf("\\n"); break;
	case '\r':   printf("\\r"); break;
	case '\t':   printf("\\t"); break;
	case '\v':   printf("\\v"); break;
	default:     printf("%o", c);  break;
	}
      }
    }
    printf("\"");
    break;

  }
  fflush(stdout);
  return TAG_T;
}


/* Evaluate an S-expression */
static lobj
l_eval(LISP l, lobj s)
{
  lobj  v, f, a, av[3];
  int   n; 

  switch(D_GET_TAG(s)){

  case TAG_NIL:        /* self-evaluating objects */
  case TAG_T:
  case TAG_INT:
  case TAG_FLOAT:
  case TAG_VECT:
  case TAG_STR:
    return s;

  case TAG_SYMB:       /* symbol ... refer to the symbol table */
    if ((v = l->t_symb_val[D_GET_DATA(s)]) == TAG_UNDEF)
      return err_msg(l, errmsg_sym_undef, 1, s);
    return v;

  case TAG_CONS:       /* cons ... function call */
    f = l_car(l,s);   /* function name or lambda exp */
    a = l_cdr(l,s);   /* actual argument list */
    if ((D_GET_TAG(f) == TAG_CONS) && (D_GET_TAG(l_car(l,f)) == TAG_SYMB) 
	&& (D_GET_DATA(l_car(l,f)) == KW_LAMBDA)){   /* lambda exp */
      if (eval_args(l, f, a, av, FTYPE_ANY_ARGS) < 0)
	return -1;
      v = apply(l, l_cdr(l,f), av[0], list_len(l, l_car(l,l_cdr(l,f))));
    } else if (D_GET_TAG(f) == TAG_SYMB){
      n = FTYPE_GET_NARGS(l->t_symb_ftype[D_GET_DATA(f)]);
      switch (FTYPE_GET_TYPE(l->t_symb_ftype[D_GET_DATA(f)])){
      case FTYPE_UNDEF:
	return err_msg(l, errmsg_func_undef, 1, f);
      case FTYPE_SPECIAL:
	v = special(l, f, a);
	break;
      case FTYPE_SYS:
	if (eval_args(l, f, a, av, n) < 0)
	  return -1;
	v = fcall(l, f, av, n);
	break;
      case FTYPE_USER:
	if (eval_args(l, f, a, av, FTYPE_ANY_ARGS) < 0)
	  return -1;
	v = apply(l, f, av[0], n);
      }
    } else {
      return err_msg(l, errmsg_ill_call, 1, s);
    }
    return v;

  default:
    fprintf(stderr, "cannot happen: l_eval()\n");
    abort();

  }
  return TAG_NIL;
}


/* Execute special form (defun, setq. etc... arguments are not evaluated) */
static lobj
special(LISP l, lobj f, lobj a)
{
  lobj  t, v, u;
  int   len, i;

  switch (D_GET_DATA(f)){

  case KW_DEFUN:
    if (list_len(l,a) < 2)
      return err_msg(l, errmsg_ill_syntax, 1, f);
    v = l_car(l,a);            /* function name  */
    if (D_GET_TAG(v) != TAG_SYMB)
      return err_msg(l, errmsg_ill_syntax, 1, f);
    t = l_cdr(l,a);            /* list of function arg and function body */
    len = list_len(l,l_car(l,t));  /* #args */
    i = D_GET_DATA(v);
    l->t_symb_fval[i]  = t;
    l->t_symb_ftype[i] = FTYPE(FTYPE_USER, len);
    break;

  case KW_SETQ:
    t = l_car(l,a);  /* symbol name */
    if (D_GET_TAG(t) != TAG_SYMB)
      return err_msg(l, errmsg_ill_type, 1, f);
    if ((v = l_eval(l, l_car(l,l_cdr(l,a)))) < 0)  /* value */
      return -1;
    l->t_symb_val[D_GET_DATA(t)] = v;
    break;

  case KW_COMMENT:
    v = TAG_T;
    break;

  case KW_QUOTE:
    v = l_car(l,a);
    break;

  case KW_PROGN:
    for (v = TAG_NIL, t = a; D_GET_TAG(t) == TAG_CONS; t = l_cdr(l,t)){
      if ((v = l_eval(l, l_car(l,t))) < 0)
	return -1;
    }
    break;

  case KW_AND:
    for (v = TAG_T, t = a; D_GET_TAG(t) == TAG_CONS; t = l_cdr(l,t)){
      if ((v = l_eval(l, l_car(l,t))) < 0)
	return -1;
      if (D_GET_TAG(v) == TAG_NIL)
	break;
    }
    break;

  case KW_OR:
    for (v = TAG_NIL, t = a; D_GET_TAG(t) == TAG_CONS; t = l_cdr(l,t)){
      if ((v = l_eval(l, l_car(l,t))) < 0)
	return -1;
      if (D_GET_TAG(v) != TAG_NIL)
	break;
    }
    break; 

  case KW_COND:
    if (D_GET_TAG(a) != TAG_CONS)
      return err_msg(l, errmsg_ill_syntax, 1, f);
    v = TAG_NIL; 
    for (t = a; D_GET_TAG(t) == TAG_CONS; t = l_cdr(l,t)){
      u = l_car(l,t);
      if (D_GET_TAG(u) != TAG_CONS)
	return err_msg(l, errmsg_ill_syntax, 1, f);
      if ((v = l_eval(l, l_car(l,u))) < 0)
	return -1;
      if (D_GET_TAG(v) != TAG_NIL){
	for (u = l_cdr(l,u); D_GET_TAG(u) == TAG_CONS; u = l_cdr(l,u)){ 
	  if ((v = l_eval(l, l_car(l,u))) < 0)
	    return -1;
	} 
	break;
      }
    }
    break;

  case KW_IF:
    if (D_GET_TAG(a) != TAG_CONS)
      return err_msg(l, errmsg_ill_syntax, 1, f);
    len = list_len(l,a);
    if ((len ==2) || (len == 3)){
      if ((v = l_eval(l, l_car(l,a))) < 0)
	return -1;
      if (D_GET_TAG(v) != TAG_NIL)
	return l_eval(l, l_car(l,l_cdr(l,a)));
      return  (len == 2) ? TAG_NIL : l_eval(l, l_car(l,l_cdr(l,l_cdr(l,a))));
    } else {
      return err_msg(l, errmsg_ill_syntax, 1, f);
    }
    break;

  }
  return v;
}


/* Evaluate arguments */
static lobj
eval_args(LISP l, lobj func, lobj arg, lobj av[3], int n)
{
  lobj  x, y;

  if ((n != FTYPE_ANY_ARGS) && (n != list_len(l,arg)))
    return err_msg(l, errmsg_ill_nargs, 1, func);

  switch (n){

  case 0:
    av[0] = TAG_NIL;
    break;

  case 1:
    if ((av[0] = l_eval(l, l_car(l,arg))) < 0)
      return -1;
    break;

  case 2:
    if ((av[0] = l_eval(l, l_car(l,arg))) < 0)
      return -1;
    if (gc_protect(l, av[0]) < 0)
      return -1;
    if ((av[1] = l_eval(l, l_car(l,l_cdr(l,arg)))) < 0)
      return -1;
    gc_unprotect(l,av[0]);
    break;

  case 3:
    if ((av[0] = l_eval(l, l_car(l,arg))) < 0)
      return -1;
    if (gc_protect(l, av[0]) < 0)
      return -1;
    if ((av[1] = l_eval(l, l_car(l,l_cdr(l,arg)))) < 0)
      return -1;
    if (gc_protect(l, av[1]) < 0)
      return -1;
    if ((av[2] = l_eval(l, l_car(l,l_cdr(l,l_cdr(l,arg))))) < 0)
      return -1;
    gc_unprotect(l, av[1]);
    gc_unprotect(l, av[0]);
    break;

  case FTYPE_ANY_ARGS:   /* return evaluated arguments as a list */
    if (D_GET_TAG(arg) != TAG_CONS){
      av[0] = TAG_NIL;
    } else {
      if ((x = l_eval(l, l_car(l,arg))) < 0)
	return -1;
      if ((av[0] = y = l_cons(l, x, TAG_NIL)) < 0)
	return -1;
      if (gc_protect(l, av[0]) < 0)
	return -1;
      for (arg = l_cdr(l,arg); D_GET_TAG(arg) == TAG_CONS; arg = l_cdr(l,arg)){
	if ((x = l_eval(l, l_car(l,arg))) < 0)
	  return -1;
	rplacd(l, y, l_cons(l, x, TAG_NIL)); 
	y = l_cdr(l,y);
      }
      gc_unprotect(l,av[0]);
    }
  }
  return av[0];
}


/* Call a built-in function */
static lobj 
fcall(LISP l, lobj f, lobj av[3], int n)
{
  lobj  t, v;
  int   r, d, i;

  switch (D_GET_DATA(f)){

  case KW_QUIT:
    quit(l);
    return TAG_NIL;
    
  case KW_EQ:
    return (av[0] == av[1]) ? TAG_T : TAG_NIL;

  case KW_CONS:
    return l_cons(l, av[0], av[1]); 

  case KW_RPLACA:
    if (D_GET_TAG(av[0]) != TAG_CONS) 
      return err_msg(l, errmsg_ill_type, 1, f);
    rplaca(l, av[0], av[1]);
    return av[1];

  case KW_RPLACD:
    if (D_GET_TAG(av[0]) != TAG_CONS) 
      return err_msg(l, errmsg_ill_type, 1, f);
    rplacd(l, av[0], av[1]);
    return av[1];

  case KW_CAR:
    if (D_GET_TAG(av[0]) != TAG_CONS) 
      return err_msg(l, errmsg_ill_type, 1, f);
    return l_car(l,av[0]);

  case KW_CDR:
    if (D_GET_TAG(av[0]) != TAG_CONS) 
      return err_msg(l, errmsg_ill_type, 1, f);
    return l_cdr(l,av[0]);

  case KW_NULL:
    return (D_GET_TAG(av[0]) == TAG_NIL)  ? TAG_T : TAG_NIL;

  case KW_SYMBP:
    return (D_GET_TAG(av[0]) == TAG_SYMB) ? TAG_T : TAG_NIL;

  case KW_NUMBERP:
    return ((D_GET_TAG(av[0]) == TAG_INT) || (D_GET_TAG(av[0]) == TAG_FLOAT)) 
      ? TAG_T : TAG_NIL;

  case KW_INTEGERP:
    return (D_GET_TAG(av[0]) == TAG_INT)    ? TAG_T : TAG_NIL;

  case KW_FLOATP:
    return (D_GET_TAG(av[0]) == TAG_FLOAT)  ? TAG_T : TAG_NIL;

  case KW_EOFP:
    return (D_GET_TAG(av[0]) == TAG_EOF)  ? TAG_T : TAG_NIL;

  case KW_LIST:
    return av[0];

  case KW_NOT:
    return (D_GET_TAG(av[0]) == TAG_NIL) ? TAG_T : TAG_NIL;

  case KW_READ:
    return l_read(l);

  case KW_READ_STR:
    if (D_GET_TAG(av[0]) != TAG_STR) 
      return err_msg(l, errmsg_ill_type, 1, f);
    return l_read_string(l, l->t_str_body[D_GET_DATA(av[0])]);

  case KW_EVAL:
    return l_eval(l, av[0]);

  case KW_PRINT:
    return l_print(l, av[0]);

  case KW_TERPRI:
    printf("\n");
    return TAG_T;

  case KW_GC:
    gcollect(l);
    return TAG_T;

  case KW_ADD:
    for (r = 0, t = av[0]; D_GET_TAG(t) == TAG_CONS; t = l_cdr(l,t)){
      if (D_GET_TAG(l_car(l,t)) != TAG_INT)
	return err_msg(l, errmsg_ill_type, 1, f);
      r = r + int_get_c(l_car(l,t));
    }
    return int_make_l(r);

  case KW_TIMES:
    for (r = 1, t = av[0]; D_GET_TAG(t) == TAG_CONS; t = l_cdr(l,t)){
      if (D_GET_TAG(l_car(l,t)) != TAG_INT)
	return err_msg(l, errmsg_ill_type, 1, f);
      r = r * int_get_c(l_car(l,t));
    }
    return int_make_l(r);

  case KW_SUB:
    if (D_GET_TAG(av[0]) == TAG_NIL){
      r = 0;
    } else if (D_GET_TAG(l_car(l,av[0])) != TAG_INT){
	return err_msg(l, errmsg_ill_type, 1, f);
    } else if (D_GET_TAG(l_cdr(l,av[0])) == TAG_NIL){
      r = 0 - int_get_c(l_car(l,av[0]));
    } else {
      r = int_get_c(l_car(l,av[0]));
      for (t = l_cdr(l,av[0]); D_GET_TAG(t) == TAG_CONS; t = l_cdr(l,t)){
	if (D_GET_TAG(l_car(l,t)) != TAG_INT)
	  return err_msg(l, errmsg_ill_type, 1, f);
	r = r - int_get_c(l_car(l,t));
      }
    }
    return int_make_l(r);

  case KW_QUOTIENT:
    if (D_GET_TAG(av[0]) == TAG_NIL){
      r = 1;
    } else if (D_GET_TAG(l_car(l,av[0])) != TAG_INT){
	return err_msg(l, errmsg_ill_type, 1, f);
    } else if ((d = int_get_c(l_car(l,av[0]))) == 0){
      return err_msg(l, errmsg_zero_div, 1, f);
    } if (D_GET_TAG(l_cdr(l,av[0])) == TAG_NIL){
      r = 1 / d;
    } else {
      for (r = d, t = l_cdr(l,av[0]); 
	   D_GET_TAG(t) == TAG_CONS; 
	   t = l_cdr(l,t)){
	if (D_GET_TAG(l_car(l,t)) != TAG_INT)
	  return err_msg(l, errmsg_ill_type, 1, f);
	if ((d = int_get_c(l_car(l,t))) == 0)
	  return err_msg(l, errmsg_zero_div, 1, f);
	r = r / d;
      }
    }
    return int_make_l(r);

  case KW_DIVIDE:
    if ((D_GET_TAG(av[0]) != TAG_INT) || (D_GET_TAG(av[1]) != TAG_INT))
      return err_msg(l, errmsg_ill_type, 1, f);
    r = int_get_c(av[0]);
    if ((d = int_get_c(av[1])) == 0)
      return err_msg(l, errmsg_zero_div, 1, f);
    return l_cons(l, int_make_l(r / d), int_make_l(r % d));

  case KW_IEQ:
    if ((D_GET_TAG(av[0]) != TAG_INT) || (D_GET_TAG(av[1]) != TAG_INT))
      return err_msg(l, errmsg_ill_type, 1, f);
    return (int_get_c(av[0]) == int_get_c(av[1])) ? TAG_T : TAG_NIL;

  case KW_IGT:
    if ((D_GET_TAG(av[0]) != TAG_INT) || (D_GET_TAG(av[1]) != TAG_INT))
      return err_msg(l, errmsg_ill_type, 1, f);
    return (int_get_c(av[0]) > int_get_c(av[1])) ? TAG_T : TAG_NIL;

  case KW_IGE:
    if ((D_GET_TAG(av[0]) != TAG_INT) || (D_GET_TAG(av[1]) != TAG_INT))
      return err_msg(l, errmsg_ill_type, 1, f);
    return (int_get_c(av[0]) >= int_get_c(av[1])) ? TAG_T : TAG_NIL;

  case KW_ILT:
    if ((D_GET_TAG(av[0]) != TAG_INT) || (D_GET_TAG(av[1]) != TAG_INT))
      return err_msg(l, errmsg_ill_type, 1, f);
    return (int_get_c(av[0]) < int_get_c(av[1])) ? TAG_T : TAG_NIL;

  case KW_ILE:
    if ((D_GET_TAG(av[0]) != TAG_INT) || (D_GET_TAG(av[1]) != TAG_INT))
      return err_msg(l, errmsg_ill_type, 1, f);
    return (int_get_c(av[0]) <= int_get_c(av[1])) ? TAG_T : TAG_NIL;
    
  case KW_VECTP:
    return (D_GET_TAG(av[0]) == TAG_VECT)  ? TAG_T : TAG_NIL;

  case KW_VECT:
    return l_list_to_vect(l,av[0]);

  case KW_MKVECT:
    if (D_GET_TAG(av[0]) != TAG_INT) 
      return err_msg(l, errmsg_ill_type, 1, f);
    return l_make_vect(l, int_get_c(av[0]));

  case KW_VECTLEN:
    if (D_GET_TAG(av[0]) != TAG_VECT) 
      return err_msg(l, errmsg_ill_type, 1, f);
    return int_make_l(l->t_vect_size[D_GET_DATA(av[0])]);
    
  case KW_VECTREF:
    if (D_GET_TAG(av[0]) != TAG_VECT) 
      return err_msg(l, errmsg_ill_type, 1, f);
    if (D_GET_TAG(av[1]) != TAG_INT) 
      return err_msg(l, errmsg_ill_type, 1, f);
    d = int_get_c(av[1]);
    if ((d < 0) || (l->t_vect_size[D_GET_DATA(av[0])] <= d))
      return err_msg(l, errmsg_ill_range, 1, f);
    return l->t_vect_body[D_GET_DATA(av[0])][d]; 

  case KW_VECTSET:
    if (D_GET_TAG(av[0]) != TAG_VECT) 
      return err_msg(l, errmsg_ill_type, 1, f);
    if (D_GET_TAG(av[1]) != TAG_INT) 
      return err_msg(l, errmsg_ill_type, 1, f);
    d = int_get_c(av[1]);
    if ((d < 0) || (l->t_vect_size[D_GET_DATA(av[0])] <= d))
      return err_msg(l, errmsg_ill_range, 1, f);
    l->t_vect_body[D_GET_DATA(av[0])][d] = av[2];
    return TAG_T; 

  case KW_STRP:
    return (D_GET_TAG(av[0]) == TAG_STR) ? TAG_T : TAG_NIL;

  case KW_MKSTR:
    if (D_GET_TAG(av[0]) != TAG_INT) 
      return err_msg(l, errmsg_ill_type, 1, f);
    d = int_get_c(av[0]);
    if (d < 0) 
      return err_msg(l, errmsg_ill_range, 1, f);
    return l_make_str(l, int_get_c(av[0]));
    
  case KW_STRLEN:
    if (D_GET_TAG(av[0]) != TAG_STR) 
      return err_msg(l, errmsg_ill_type, 1, f);
    return int_make_l(l->t_str_size[D_GET_DATA(av[0])]);

  case KW_STRAPPND: 
    if (D_GET_TAG(av[0]) == TAG_NIL)
      return l_make_str(l,0); 
    for (r = 0, t = av[0]; D_GET_TAG(t) == TAG_CONS; t = l_cdr(l,t)){
      if (D_GET_TAG(l_car(l,t)) != TAG_STR)
	return err_msg(l, errmsg_ill_type, 1, f);
      r += l->t_str_size[D_GET_DATA(l_car(l,t))];
    }
    v = l_make_str(l,r);
    for (r = 0, t = av[0]; D_GET_TAG(t) == TAG_CONS; t = l_cdr(l,t)){
      for (i = 0; i < l->t_str_size[D_GET_DATA(t)]; i++)
	l->t_str_body[D_GET_DATA(v)][r+i] 
	  = l->t_str_body[D_GET_DATA(l_car(l,t))][i];
      r += l->t_str_size[D_GET_DATA(l_car(l,t))];
    }
    return v;

  case KW_EQUAL:
    return l_equal(l, av[0], av[1]);

  default:
    fprintf(stderr, "cannot happen: fcall()\n");
    abort();
  }

  return TAG_NIL;
}


/* Function application (user defined function) */
static lobj
apply(LISP l, lobj func, lobj aparams, int n) 
{
  lobj  fdef, fbody, f, sym, a, v;
  int   i;

  if (l->t_stack_ptr + n + 1 >= l->t_stack_size){   /* stack overflow */
    /* We do not extend stack.  
     * Probably it is a bug, such as infinite loop of function call.
     */
    return err_msg(l, errmsg_stack_of, 0, 0);
  }

  if (D_GET_TAG(func) == TAG_SYMB){         /* function symbol */
    fdef = l->t_symb_fval[D_GET_DATA(func)];
  } else if (D_GET_TAG(func) == TAG_CONS){  /* lambda exp */
    fdef = func;
  }

  /* bind */
  f = l_car(l,fdef);  /* formal parameters */
  a = aparams;      /* actual parameters */
  l->t_stack_ptr = l->t_stack_ptr + n;
  for (i = 0; i < n; i++, f = l_cdr(l,f), a = l_cdr(l,a)){
    sym = l_car(l,f);
    /* push old symbol values onto stack */
    l->t_stack[l->t_stack_ptr - i - 1] = l->t_symb_val[D_GET_DATA(sym)];
    /* bind argument value to symbol */
    l->t_symb_val[D_GET_DATA(sym)] = l_car(l,a);
  }

  if (gc_protect(l,aparams) < 0)
    return -1;

  /* evaluate function body */
  fbody = l_cdr(l,fdef);  /* function body */
  for (v = TAG_NIL; D_GET_TAG(fbody) == TAG_CONS; fbody = l_cdr(l,fbody)){
    if ((v = l_eval(l, l_car(l,fbody))) < 0)
      break;   /* error ... never return immediately - need unbinding. */
  }

  /* pop gc_protected objects, including 'gc_unprotect(aparams)'. */
  while ((l->t_stack[l->t_stack_ptr-1] & D_GC_MARK) != 0)
    --l->t_stack_ptr;   

  /* unbind: restore old variable values from stack */
  for (i = 0, f = l_car(l,fdef); i < n; i++, f = l_cdr(l,f)){
    sym = l_car(l,f);
    l->t_symb_val[D_GET_DATA(sym)] = l->t_stack[l->t_stack_ptr - i - 1];
  }
  l->t_stack_ptr = l->t_stack_ptr - n;

  return v;
}


/* Print an error message */
static int
err_msg(LISP l, char *msg, int f, lobj s)
{
  printf("\nERROR. \n%s", msg);
  if (f != 0)
    l_print(l, s);
  printf("\n");
  return -1;
}

/* Length of a list */
static int
list_len(LISP l, lobj s)
{
  int  i;

  for (i = 0; D_GET_TAG(s) == TAG_CONS; s = l_cdr(l,s))
    i++;
  return i;
}

/* "Replace car" opration ... rewite car part of a cons cell */
static void
rplaca(LISP l, lobj s, lobj car)
{
  l->t_cons_car[D_GET_DATA(s)] = car;
}

/* "Replace cdr" opration ... rewite cdr part of a cons cell */
static void
rplacd(LISP l, lobj s, lobj cdr)
{
  l->t_cons_cdr[D_GET_DATA(s)] = cdr;
}


/* "Cons" opration */
static lobj
l_cons(LISP l, lobj car, lobj cdr)
{
  int  s;

  if (l->t_cons_free < 0){   /*  no cons cells */
    if (gc_protect(l, car) < 0)
      return -1;
    if (gc_protect(l, cdr) < 0)
      return -1;
    gcollect(l);           /* invoke garbage collector */
    gc_unprotect(l, cdr);
    gc_unprotect(l, car);
  }

  /* get a free cons cell from a free list */
  s = l->t_cons_free;
  if (l->t_cons_car[l->t_cons_free] != l->t_cons_free)
    l->t_cons_free  = l->t_cons_car[l->t_cons_free];  /* next free cell */
  else
    l->t_cons_free = -1;                        /* self-loop: exhausted */

  /* constract a new cell */
  l->t_cons_car[s] = car;
  l->t_cons_cdr[s] = cdr;

  return (TAG_CONS | s);
}

/* "Car" opration */
static lobj
l_car(LISP l, lobj s)
{
  return l->t_cons_car[D_GET_DATA(s)];
}

/* "Cdr" opration */
static lobj
l_cdr(LISP l, lobj s)
{
  return l->t_cons_cdr[D_GET_DATA(s)];
}

static lobj
l_equal(LISP l, lobj s1, lobj s2)
{
  int  d1 = s1 & D_MASK_DATA;
  int  d2 = s2 & D_MASK_DATA;
  int  i;

  if (D_GET_TAG(s1) != D_GET_TAG(s2))
    return TAG_NIL;
  switch (D_GET_TAG(s1)){
  case TAG_CONS:
    if (l_equal(l, l_car(l,s1), l_car(l,s1)) == TAG_NIL)
      return TAG_NIL;
    return l_equal(l, l_car(l,s2), l_car(l,s2));
  case TAG_STR:
    if (l->t_str_size[d1] != l->t_str_size[d2])
      return TAG_NIL;
    if (strcmp(l->t_str_body[d1], l->t_str_body[d2]) == 0) 
      return TAG_T;
    return TAG_NIL;
  case TAG_VECT:
    if (l->t_vect_size[d1] != l->t_vect_size[d2])
      return TAG_NIL;
    for (i = 0; i < l->t_vect_size[d1]; i++){
      if (l_equal(l, l->t_vect_body[d1][i], l->t_vect_body[d2][i]) == TAG_NIL)
	return TAG_NIL;
    }
    return TAG_T;
  default:
    return (s1 == s2) ? TAG_T : TAG_NIL;
  }
}

static lobj
l_make_vect(LISP l, int n)
{
  int  vi, i;

  if (l->t_vect_free < 0)
    gcollect(l);           /* invoke garbage collector */

  vi = l->t_vect_free;
  if (l->t_vect_free != l->t_vect_size[l->t_vect_free]){
    l->t_vect_free = l->t_vect_size[l->t_vect_free];  /* next free slot */
  } else {
    l->t_vect_free = -1;                              /* exhausted */
  }

  l->t_vect_size[vi] = n;
  l->t_vect_body[vi] = (lobj*) calloc(n, sizeof(lobj));
  for (i = 0; i < n; i++)
    l->t_vect_body[vi][i] = TAG_NIL;

  return (TAG_VECT | vi);
}

static lobj
l_list_to_vect(LISP l, lobj s)
{
  lobj  v;
  int   len, i;

  len = list_len(l, s);

  if (gc_protect(l, s) < 0)
    return -1;
  v = l_make_vect(l, len);
  gc_unprotect(l, s);

  for (i = 0; i < len; i++, s = l_cdr(l,s))
    l->t_vect_body[D_GET_DATA(v)][i] = l_car(l,s);

  return  v;
}

static lobj
l_make_str(LISP l, int n)
{
  int  vi, i;

  if (l->t_str_free < 0)
    gcollect(l);           /* invoke garbage collector */

  vi = l->t_str_free;
  if (l->t_str_free != l->t_str_size[l->t_str_free]){
    l->t_str_free = l->t_str_size[l->t_str_free];   /* next free slot */
  } else {
    l->t_str_free = -1;                             /* exhausted */
  }

  l->t_str_size[vi] = n;
  l->t_str_body[vi] = (char*) malloc(n+1);
  for (i = 0; i < n; i++)
    l->t_str_body[vi][i] = '?';
  l->t_str_body[vi][n] = '\0';

  return (TAG_STR | vi);
}


/* Quit micro lisp */
static void
quit(LISP l)
{
  printf("\nBYE\n");
  exit(0);
}


/* Garbage collector */
static void
gcollect(LISP l)
{
  int  n_cells, n_vects, n_strs, n_floats, n, i, p;

  /* mark */
  for (i = 0; i < l->t_symb_free; i++)
    gc_mark(l, l->t_symb_val[i]);
  for (i = 0; i < l->t_symb_free; i++)
    gc_mark(l, l->t_symb_fval[i]);
  for (i = 0; i < l->t_stack_ptr; i++)
    gc_mark(l, l->t_stack[i]);
#ifndef CAMPUS_LISP
  for (i = 0; i < vf_table_size_alloced(l->l_extern_gc_protect); i++){
    if (vf_table_id_occupiedp(l->l_extern_gc_protect, i))
      gc_mark(l, (lobj)vf_table_get_obj(l->l_extern_gc_protect, i));
  }
#endif

  /* sweep cell space */
  n_cells = 0;
  l->t_cons_free = p = -1;
  for (i = 0; i < l->t_cons_ncons; i++){
    if ((l->t_cons_car[i] & D_GC_MARK) == 0){   /* not referenced; collect */
      n_cells++;
      if (l->t_cons_free == -1){
	l->t_cons_free = i;
      } else {
	l->t_cons_car[p] = i; 
      }
      l->t_cons_car[i] = i; 
      p = i;
    }
    l->t_cons_car[i] &= ~D_GC_MARK;   /* clear gc mark */
  }
  if (n_cells == 0){    
    /* extend cell space */
    n = l->t_cons_ncons + NCONS ;
    l->t_cons_car  = (int*) realloc(l->t_cons_car, n * sizeof(int));
    l->t_cons_cdr  = (int*) realloc(l->t_cons_cdr, n * sizeof(int));
    l->t_cons_free = l->t_cons_ncons;
    for (i = l->t_cons_ncons; i < n-1; i++)
      l->t_cons_car[i] = i + 1;
    l->t_cons_car[n-1] = n-1;
    l->t_cons_ncons = n;
  }

  /* sweep vector space */
  n_vects = 0;
  l->t_vect_free = p = -1;
  for (i = 0; i < l->t_vect_nvects; i++){
    if (l->t_vect_gcmark[i] == 0){    /* not referenced; collect */
      n_vects++;
      if (l->t_vect_free == -1){
	l->t_vect_free = i;
      } else {
	l->t_vect_size[p] = i; 
      }
      l->t_vect_size[i] = i; 
      p = i;
      if (l->t_vect_body[i] != NULL){
	free(l->t_vect_body[i]);
	l->t_vect_body[i] = NULL;
      }
    }
    l->t_vect_gcmark[i] = 0;   /* clear gc mark */
  }
  if (n_vects == 0){
    /* extend vector space */
    n = l->t_vect_nvects + NVECTS;
    l->t_vect_size   = (int*)  realloc(l->t_vect_size,   n * sizeof(int));
    l->t_vect_body   = (lobj**)realloc(l->t_vect_body,   n * sizeof(lobj));
    l->t_vect_gcmark = (char*) realloc(l->t_vect_gcmark, n * sizeof(char*));
    for (i = l->t_vect_nvects; i < n; i++){
      l->t_vect_body[i] = NULL;
      l->t_vect_gcmark[i] = 0;
    }
    /* free list */
    l->t_vect_free = l->t_vect_nvects;
    for (i = l->t_vect_nvects; i < n-1; i++)
      l->t_vect_size[i] = i+1;
    l->t_vect_size[n-1] = n-1;
    l->t_vect_nvects = n;
  }

  /* sweep float space */
  n_floats = 0;
  l->t_flo_free = p = -1;
  for (i = 0; i < l->t_flo_nfloats; i++){
    if (l->t_flo_gcmark[i] == 0){    /* not referenced; collect */
      n_floats++;
      if (l->t_flo_free == -1){
	l->t_flo_free = i;
      } else {
	l->t_flo_body[p].next = i; 
      }
      l->t_flo_body[i].next = i; 
      p = i;
    }
    l->t_flo_gcmark[i] = 0;   /* clear gc mark */
  }
  if (n_floats == 0){
    /* extend vector space */
    n = l->t_flo_nfloats + NFLOATS;
    l->t_flo_body   = realloc(l->t_flo_body,   n * sizeof(double));
    l->t_flo_gcmark = realloc(l->t_flo_gcmark, n * sizeof(char*));
    for (i = l->t_flo_nfloats; i < n; i++)
      l->t_flo_gcmark[i] = 0;
    /* free list */
    l->t_flo_free = l->t_flo_nfloats;
    for (i = l->t_flo_nfloats; i < n-1; i++)
      l->t_flo_body[i].next = i+1;
    l->t_flo_body[n-1].next = n-1;
    l->t_flo_nfloats = n;
  }

  /* sweep string space */
  n_strs = 0;
  l->t_str_free = p = -1;
  for (i = 0; i < l->t_str_nstrs; i++){
    if (l->t_str_gcmark[i] == 0){    /* not referenced; collect */
      n_strs++;
      if (l->t_str_free == -1){
	l->t_str_free = i;
      } else {
	l->t_str_size[p] = i; 
      }
      l->t_str_size[i] = i; 
      p = i;
      if (l->t_str_body[i] != NULL){
	free(l->t_str_body[i]);
	l->t_str_body[i] = NULL;
      }
    }
    l->t_str_gcmark[i] = 0;   /* clear gc mark */
  }
  if (n_strs == 0){
    /* extend string space */
    n = l->t_str_nstrs + NSTRS;
    l->t_str_size   = (int*)  realloc(l->t_str_size,   n * sizeof(int));
    l->t_str_body   = (char**)realloc(l->t_str_body,   n * sizeof(lobj));
    l->t_str_gcmark = (char*) realloc(l->t_str_gcmark, n * sizeof(char*));
    for (i = l->t_str_nstrs; i < n; i++){
      l->t_str_body[i] = NULL;
      l->t_str_gcmark[i] = 0;
    }
    /* free list */
    l->t_str_free = l->t_str_nstrs;
    for (i = l->t_str_nstrs; i < n-1; i++)
      l->t_str_size[i] = i + 1;
    l->t_str_size[n-1] = n-1;
    l->t_str_nstrs = n;
  }
}

/* mark recursively */
static void
gc_mark(LISP l, lobj s)
{
  int  i;

  if (D_GET_TAG(s) == TAG_CONS){
    for ( ; D_GET_TAG(s) == TAG_CONS; s = l_cdr(l,s)){
      if ((l->t_cons_car[D_GET_DATA(s)] & D_GC_MARK) != 0) /* visited before */
	return;
      l->t_cons_car[D_GET_DATA(s)] |= D_GC_MARK;  /* mark */
      gc_mark(l, l_car(l,s));                       /* visit car part */
    }
    gc_mark(l, s);   /* s can be a vector */
  } else if (D_GET_TAG(s) == TAG_VECT){
    if (l->t_vect_gcmark[D_GET_DATA(s)] != 0)   /* visited before */
      return;
    l->t_vect_gcmark[D_GET_DATA(s)] = 1;        /* mark */
    for (i = 0; i < l->t_vect_size[D_GET_DATA(s)]; i++)   /* visit elements */
      gc_mark(l, l->t_vect_body[D_GET_DATA(s)][i]);
  } else if (D_GET_TAG(s) == TAG_STR){
    if (l->t_str_gcmark[D_GET_DATA(s)] != 0)    /* visited before */
      return;
    l->t_str_gcmark[D_GET_DATA(s)] = 1;         /* mark */
  } else if (D_GET_TAG(s) == TAG_FLOAT){
    l->t_flo_gcmark[D_GET_DATA(s)] = 1;
  }
}

/* protect/unprotect temporary objects from garbage collector */
static int
gc_protect(LISP l, lobj s)
{
  if (D_GET_TAG(s) == TAG_CONS){  /* save only cons cells */
    if (l->t_stack_ptr >= l->t_stack_size)     /* stack overflow */
      return err_msg(l, errmsg_stack_of, 0, 0);
    l->t_stack[l->t_stack_ptr++] = (D_GC_MARK | s);  
  }
  return 0;
}

static void
gc_unprotect(LISP l, lobj s)
{
  if (D_GET_TAG(s) == TAG_CONS)
    l->t_stack_ptr = l->t_stack_ptr - 1;
}


/* Make a Lisp integer from a C integer */
static lobj 
int_make_l(int v)
{
  return (TAG_INT | ((unsigned int)v & D_MASK_DATA));
}

/* Make a C integer from a Lisp integer */
static int 
int_get_c(lobj s)
{
  if (((unsigned int)s & D_INT_SIGN_BIT) == 0)
    return ((unsigned int)s & D_MASK_DATA);
  return (int) ((unsigned int)s | ~D_MASK_DATA);
}


static lobj
l_make_float(LISP l, double val)
{
  int  vi;

  if (l->t_flo_free < 0)
    gcollect(l);           /* invoke garbage collector */

  vi = l->t_flo_free;
  if (l->t_flo_free != l->t_flo_body[l->t_vect_free].next){
    l->t_flo_free = l->t_flo_body[l->t_flo_free].next ;  /* next free slot */
  } else {
    l->t_flo_free = -1;                                  /* exhausted */
  }
  l->t_flo_body[vi].v = val;

  return (TAG_FLOAT | vi);
}

static double
float_get_c(LISP l, lobj s)
{
  int  vi;

  vi = (unsigned int)s & D_MASK_DATA;
  return l->t_flo_body[vi].v;
}


/* Make a new symbol */
static lobj
symb_make(LISP l, char *p)
{
  int  s, n, i;

  for (i = 0; i < l->t_symb_nsymbs-1; i++){
    if ((l->t_symb_pname[i] >= 0)
	&& (strcmp(p, &l->t_pnames[l->t_symb_pname[i]]) == 0))
      return i;
  }

  s = l->t_symb_free;
  l->t_symb_free++;
  if (l->t_symb_free == l->t_symb_nsymbs){
    n = l->t_symb_nsymbs + NSYMBS;
    l->t_symb_pname = (int*)  realloc(l->t_symb_pname, n * sizeof(int));
    l->t_symb_val   = (lobj*) realloc(l->t_symb_val,   n * sizeof(lobj));
    l->t_symb_fval  = (lobj*) realloc(l->t_symb_fval,  n * sizeof(lobj));
    l->t_symb_ftype = (int*)  realloc(l->t_symb_ftype, n * sizeof(int));
    for (i = l->t_symb_nsymbs; i < n; i++)
      l->t_symb_pname[i] = -1;
    l->t_symb_nsymbs = n;
  }

  l->t_symb_pname[s] = l->t_pnames_free;
  l->t_symb_val[s]   = TAG_UNDEF;   /* undefined value */
  l->t_symb_fval[s]  = TAG_UNDEF;   /* undefined value */
  l->t_symb_ftype[s] = FTYPE_UNDEF; /* undefined type */

  do {
    if (l->t_pnames_free == l->t_pnames_size){
      l->t_pnames = (char*) realloc(l->t_pnames, 
				    (l->t_pnames_size+PNAME_SIZE));
      l->t_pnames_size += PNAME_SIZE;
    }
    l->t_pnames[l->t_pnames_free++] = tolower(*(p++));
  } while (*p != '\0');
  l->t_pnames[l->t_pnames_free++] = '\0';

  return (TAG_SYMB | s);
}



#ifndef CAMPUS_LISP

#define L(eng)  ((LISP)((eng)->lisp))

/*
 * Interface Functions
 */

void
vf_lisp_engine_new(VF_Engine eng)
{
  if ((eng == NULL) && (eng->lisp != NULL))
    vf_fatal_errmsg(eng, "vf_lisp_init");

  eng->lisp = vf_malloc(eng, sizeof(struct s_lisp));
  init(eng->lisp);
  vf_table_new(eng, &L(eng)->l_extern_gc_protect, 
	       "lisp-external-gc-protect");
}

void
vf_lisp_engine_delete(VF_Engine eng)
{
  if ((eng == NULL) && (eng->lisp == NULL))
    vf_fatal_errmsg(eng, "vf_lisp_finish");

  vf_table_delete(&L(eng)->l_extern_gc_protect);
  finish(eng->lisp);
  vf_free(eng, eng->lisp);
  eng->lisp = NULL;
}

void
vf_lisp_toplevel(VF_Engine eng)
{
  printf("CAMPUS LIsP 2.0 (Lemon version)\n");
  printf("Type ^D to exit\n");
  toplevel(eng->lisp);
  printf("\nBYE\n");
}

vf_sexp
vf_lisp_read_file(VF_Engine eng, FILE *fp)
{
  return l_read_fp(eng->lisp, fp);
}

vf_sexp
vf_lisp_read_string(VF_Engine eng, char *str)
{
  return l_read_string(eng->lisp, str);
}

vf_sexp
vf_lisp_intern(VF_Engine eng, char *name)
{
  return symb_make(eng->lisp, name);
}

void
vf_lisp_gc(VF_Engine eng)
{
  gcollect(eng->lisp);
}

int
vf_lisp_gc_protect(VF_Engine eng, vf_sexp s)
{
  int  gcpid;

  if (eng == NULL)
    vf_fatal_errmsg(eng, "vf_lisp_gc_protect");

  gcpid = vf_table_put(L(eng)->l_extern_gc_protect, (void*)s);

  return  gcpid;
}

void
vf_lisp_gc_unprotect(VF_Engine eng, int gcpid)
{
  if (eng == NULL)
    vf_fatal_errmsg(eng, "vf_lisp_gc_unprotect");
  if (!vf_table_id_occupiedp(L(eng)->l_extern_gc_protect, gcpid))
    vf_fatal_errmsg(eng, "vf_lisp_gc_unprotect");

  vf_table_remove(L(eng)->l_extern_gc_protect, gcpid);
}

void
vf_lisp_gc_unprotect_all(VF_Engine eng)
{
  int gcpid, n;

  if (eng == NULL)
    vf_fatal_errmsg(eng, "vf_lisp_gc_unprotect_all");

  n = vf_table_size_alloced(L(eng)->l_extern_gc_protect);
  for (gcpid = 0; gcpid < n; gcpid++){
    if (vf_table_id_occupiedp(L(eng)->l_extern_gc_protect, gcpid)){
      vf_table_remove(L(eng)->l_extern_gc_protect, gcpid);
    }
  }
}

vf_sexp
vf_lisp_setq(VF_Engine eng, vf_sexp var, vf_sexp val)
{
  if (D_GET_TAG(var) != TAG_SYMB)
    vf_fatal_errmsg(NULL, "vf_lisp_setq");
  L(eng)->t_symb_val[D_GET_DATA(var)] = val;
  return val;
}

vf_sexp
vf_lisp_symb_val(VF_Engine eng, vf_sexp var)
{
  if (D_GET_TAG(var) != TAG_SYMB)
    vf_fatal_errmsg(NULL, "vf_lisp_symb_val");
  return L(eng)->t_symb_val[D_GET_DATA(var)];
}

vf_sexp
vf_lisp_eval(VF_Engine eng, vf_sexp s)
{
  return l_eval(eng->lisp, s);
}

vf_sexp
vf_lisp_print(VF_Engine eng, vf_sexp s)
{
  return l_print(eng->lisp, s);
}

vf_sexp
vf_lisp_terpri(VF_Engine eng)
{
  return printf("\n");
}

int
vf_lisp_null(VF_Engine eng, vf_sexp s)
{
  return (s == TAG_NIL);
}

int
vf_lisp_nilp(VF_Engine eng, vf_sexp s)
{
  return (s == TAG_NIL);
}

int
vf_lisp_tp(VF_Engine eng, vf_sexp s)
{
  return (s == TAG_T);
}

int
vf_lisp_eofp(VF_Engine eng, vf_sexp s)
{
  return (s == TAG_EOF);
}

int
vf_lisp_symbolp(VF_Engine eng, vf_sexp s)
{
  return (D_GET_TAG(s) == TAG_SYMB);
}

int
vf_lisp_numberp(VF_Engine eng, vf_sexp s)
{
  return (D_GET_TAG(s) == TAG_INT) || (D_GET_TAG(s) == TAG_FLOAT) ;
}

int
vf_lisp_integerp(VF_Engine eng, vf_sexp s)
{
  return (D_GET_TAG(s) == TAG_INT);
}

int
vf_lisp_floatp(VF_Engine eng, vf_sexp s)
{
  return (D_GET_TAG(s) == TAG_FLOAT) ;
}

int
vf_lisp_consp(VF_Engine eng, vf_sexp s)
{
  return (D_GET_TAG(s) == TAG_CONS);
}

int
vf_lisp_vectorp(VF_Engine eng, vf_sexp s)
{
  return (D_GET_TAG(s) == TAG_VECT);
}

int
vf_lisp_stringp(VF_Engine eng, vf_sexp s)
{
  return (D_GET_TAG(s) == TAG_STR);
}

int
vf_lisp_listp(VF_Engine eng, vf_sexp s)
{
  for ( ; D_GET_TAG(s) == TAG_CONS; s = l_cdr(eng->lisp, s))
    ;
  return (s == TAG_NIL);
}

int
vf_lisp_alistp(VF_Engine eng, vf_sexp s)
{
  vf_sexp  v;

  for ( ; D_GET_TAG(s) == TAG_CONS; s = l_cdr(eng->lisp, s)){
    v = vf_lisp_car(eng,s);
    if (!vf_lisp_listp(eng, v))
      return VF_FALSE;
    if (!vf_lisp_symbolp(eng, vf_lisp_car(eng,v)) 
	&& !vf_lisp_stringp(eng, vf_lisp_car(eng,v)))
      return VF_FALSE;
  }
  return (s == TAG_NIL);
}

vf_sexp
vf_lisp_assoc(VF_Engine eng, char *name, vf_sexp s)
{
  vf_sexp  v;

  if (!vf_lisp_alistp(eng, s))
    return TAG_NIL;
  for ( ; D_GET_TAG(s) == TAG_CONS; s = l_cdr(eng->lisp, s)){
    v = vf_lisp_caar(eng,s);
    if (strcmp(name, vf_lisp_get_cstring(eng, v)) == 0)
      return vf_lisp_cdar(eng,s);
  }
  return TAG_NIL;
}

int
vf_lisp_length(VF_Engine eng, vf_sexp s)
{
  return list_len(eng->lisp, s);
}

int
vf_lisp_eq(VF_Engine eng, vf_sexp s1, vf_sexp s2)
{
  return (s1 == s2);
}

int
vf_lisp_equal(VF_Engine eng, vf_sexp s1, vf_sexp s2)
{
  return (l_equal(eng->lisp, s1, s2) != TAG_NIL);
}

char*
vf_lisp_get_cstring(VF_Engine eng, vf_sexp s)
{
  if ((D_GET_TAG(s) != TAG_SYMB) && (D_GET_TAG(s) != TAG_STR))
    vf_fatal_errmsg(NULL, "vf_lisp_get_cstring");
  if (D_GET_TAG(s) == TAG_SYMB)
    return &(L(eng)->t_pnames[L(eng)->t_symb_pname[s & D_MASK_DATA]]);
  return L(eng)->t_str_body[s & D_MASK_DATA];
}

int
vf_lisp_get_cint(VF_Engine eng, vf_sexp s)
{
  if (D_GET_TAG(s) != TAG_INT)
    vf_fatal_errmsg(NULL, "vf_lisp_get_cint");
  return int_get_c(s);
}

double
vf_lisp_get_cnumber(VF_Engine eng, vf_sexp s)
{
  if (D_GET_TAG(s) == TAG_INT)
    return (double)int_get_c(s);
  if (D_GET_TAG(s) == TAG_FLOAT)
    return (double)float_get_c(eng->lisp, s);
  vf_fatal_errmsg(NULL, "vf_lisp_get_cint");
  return 0.0;
}

vf_sexp
vf_lisp_nil(VF_Engine eng)
{
  return TAG_NIL;
}

vf_sexp
vf_lisp_cons(VF_Engine eng, vf_sexp s1,vf_sexp s2)
{
  return l_cons(eng->lisp, s1, s2);
}

vf_sexp
vf_lisp_car(VF_Engine eng, vf_sexp s)
{
  if (D_GET_TAG(s) != TAG_CONS)
    vf_fatal_errmsg(NULL, "vf_lisp_car");
  return l_car(eng->lisp, s);
}

vf_sexp
vf_lisp_cdr(VF_Engine eng, vf_sexp s)
{
  if (D_GET_TAG(s) != TAG_CONS)
    vf_fatal_errmsg(NULL, "vf_lisp_cdr");
  return l_cdr(eng->lisp, s);
}

vf_sexp
vf_lisp_caar(VF_Engine eng, vf_sexp s)
{
  return l_car(eng->lisp, l_car(eng->lisp, s));
}

vf_sexp
vf_lisp_cadr(VF_Engine eng, vf_sexp s)
{
  return l_car(eng->lisp, l_cdr(eng->lisp, s));
}

vf_sexp
vf_lisp_cdar(VF_Engine eng, vf_sexp s)
{
  return l_cdr(eng->lisp, l_car(eng->lisp, s));
}

vf_sexp
vf_lisp_cddr(VF_Engine eng, vf_sexp s)
{
  return l_cdr(eng->lisp, l_cdr(eng->lisp, s));
}

void
vf_lisp_rplaca(VF_Engine eng, vf_sexp s, vf_sexp car)
{
  rplaca(eng->lisp, s, car);
}

void
vf_lisp_rplacd(VF_Engine eng, vf_sexp s,vf_sexp cdr)
{
  rplacd(eng->lisp, s, cdr);
}

int
vf_lisp_vector_length(VF_Engine eng, vf_sexp s)
{
  if (D_GET_TAG(s) != TAG_VECT) 
    vf_fatal_errmsg(NULL, "vf_lisp_vector_length");
  return (D_GET_TAG(s) == TAG_VECT);
}

vf_sexp
vf_lisp_vector_ref(VF_Engine eng, vf_sexp s, int i)
{
  if (D_GET_TAG(s) != TAG_VECT) 
    vf_fatal_errmsg(NULL, "vf_lisp_vector_ref");
  if ((i < 0) || (L(eng)->t_vect_size[D_GET_DATA(s)] <= i))
    vf_fatal_errmsg(NULL, "vf_lisp_vector_ref");
  return L(eng)->t_vect_body[D_GET_DATA(s)][i]; 
}

int
vf_lisp_string_length(VF_Engine eng, vf_sexp s)
{
  if (D_GET_TAG(s) != TAG_STR) 
    vf_fatal_errmsg(NULL, "vf_lisp_string_length");
  return L(eng)->t_vect_size[D_GET_DATA(s)];
}

char*
vf_lisp_string(VF_Engine eng, vf_sexp s)
{
  if (D_GET_TAG(s) != TAG_STR)
    vf_fatal_errmsg(NULL, "vf_lisp_string");
  return L(eng)->t_str_body[D_GET_DATA(s)];
}

#ifdef DEBUG
int
main(int argc, char **argv)
{
  VF_Engine  eng;
  vf_sexp    s, v;
  int        p;

  argc--; argv++;
  if (argc == 0){
    printf("vflisp [S-EXP ...]\n");
    exit(0);
  }

  vf_engine_new(&eng, "vflisp");
  vf_lisp_engine_new(eng);

  for ( ; argc > 0; argc--, argv++){
    s = vf_lisp_read_string(eng, argv[0]);
    vf_lisp_print(eng, s);
    vf_lisp_terpri(eng);

    p = vf_lisp_gc_protect(eng, s);
    v = vf_lisp_eval(eng, s); 
    vf_lisp_gc_unprotect(eng, p);

    if (v < 0)
      continue;

    printf("  ==> ");
    vf_lisp_print(eng, v);
    vf_lisp_terpri(eng);
  }

  vf_engine_delete(&eng);

  return 0;
}

#endif /*DEBUG*/ 
#endif /*CAMPUS_LISP*/

/* END */
